/*	$Id: walkMainCodels.cc 2008/04/04 19:22:05 mallet $ */

/*
 * Copyright (c) 2007 Mathieu Poirier                 --  Thu Apr  5 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: walkMainCodels.cc 2008/04/04 19:22:05 mallet $";

#include <err.h>
#include <iostream>

#include <portLib.h>

#include "server/walkHeader.h"

#include "KineoWorks2/kwsConfig.h"
#include "KineoKCDModel/kppKCDPolyhedron.h"

#include "hppWalkPlanner/hppWalkPlanner.h"
#include "hppOpenHRP/configSwap.h"
#include "KineoModel/kppLicense.h"
#include "KineoModel/kppJointComponent.h"
#include "hppCorbaServer/hppciOpenHrp.h"

#include "genom-openhrp/genom-hrp2.h"
#include "netshring.h"

#include "hrp2_14/hrp2_14.h"

/* --- local data ---------------------------------------------------- */

static netshring_id_t shtarget;		/**< shared trajectory buffer */
static ChppWalkPlanner *hppPlanner;	/**< hpp planner */


/* --- walkMainInit -------------------------------------------------- */

/** Initialization codel (fIDS, ...)
 *
 * \returns OK or ERROR
 */

STATUS
walkMainInit(int *report)
{
  SDI_F->initialized = 0;

  /* create shared trajectory buffer */
  shtarget = netshring_create(WALK_TARGET_CONFIG_NAME);
  if (!shtarget) {
    warnx("cannot create shared buffer");
    return ERROR;
  }

  return OK;
}


/* --- walkMainEnd --------------------------------------------------- */

/** Termination codel
 *
 * \returns OK or ERROR
 */

STATUS
walkMainEnd(void)
{
  /* delete shared trajectory buffer */
  if (shtarget) {
    netshring_delete(shtarget);
    shtarget = NULL;
  }

  return OK;
}


/* --- Init ---------------------------------------------------------- */

/** walkInitMain  -  codel EXEC of Init
 *
 * Reports:      OK
 *              S_walk_NAMESERVER_NOT_FOUND
 *              S_walk_OPENHRP_NOT_FOUND
 *              S_walk_CORBA_NOT_FOUND
 *              S_walk_KINEOWORKS_LICENSE_NOT_FOUND
 *              S_walk_MODELLOADERJRL_NOT_FOUND
 *              S_walk_ALREADY_INITIALIZED
 *              S_walk_COULD_NOT_BUILD_HRP2
 *              S_walk_COULD_NOT_BUILD_OBSTACLE
 *
 * \returns EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkInitMain(WALK_XY_ROBOT_BOUNDS_STR* robotBounds, int *report)
{
  // ----------- 1. init KineoWorks  -------------------------------

  if(!CkppLicense::initialize()) {
    cerr << "License for KPP SDK NOT FOUND" << endl;
    *report = S_walk_KINEOWORKS_LICENSE_NOT_FOUND;
    return ETHER;
  }
  cerr << "License for KPP SDK found" << endl;

  // walk planner
  hppPlanner = new ChppWalkPlanner();
  if (hppPlanner == NULL ) {
    cerr << "COULD NOT CREATE a hppWalkPlanner" << endl;
    *report = S_walk_HPP_PLANNER_NOT_FOUND;
    return ETHER;
  }

  // ---------- 2. Load Robot Model from ModelLoader ------------------

  ChppciOpenHrpClient openHrpCorbaClient(hppPlanner);
  if (openHrpCorbaClient.loadHrp2Model() != KD_OK) {
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }


  // ---------- 3. Initialize the walkPlanner with the devices -------

  CkppDeviceComponentShPtr hrp2 = hppPlanner->robotIthProblem(0);
  CkwsConfigShPtr halfSittingConfig = CkwsConfig::create(hrp2) ;
  double dInitPos[46] = HALFSITTINGPOSITION_RAD_KINEO ;
  std::vector<double>  halfSittingVector (dInitPos  , dInitPos + sizeof(dInitPos) / sizeof(double) );
  halfSittingConfig->setDofValues( halfSittingVector ) ;
  hrp2->setCurrentConfig(*halfSittingConfig);

  // Create bounding box robot and place it at beginning of problem vector.
  hppPlanner->initializeProblem();

  // Initialize x-y bounds of the bounding box of the robot
  CkppDeviceComponentShPtr boundingBox = hppPlanner->robotIthProblem(0);
  if (!boundingBox) {
    cerr << "walk::Init: there is no robot in hppPlanner." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  CkppJointComponentShPtr kppJoint = boundingBox->rootJointComponent();
  if (!kppJoint) {
    cerr << "walk::Init: bounding box robot has no root joint." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  CkwsJointShPtr kwsJoint = kppJoint->kwsJoint();
  if (kwsJoint->countDofs() != 6) {
    cerr << "walk::Init: Root joint of bounding box is not freeflyer." << endl;
    *report = S_walk_COULD_NOT_BUILD_ROBOT;
    return ETHER;
  }

  kwsJoint->dof(0)->isBounded(true);
  kwsJoint->dof(0)->bounds(robotBounds->xMin, robotBounds->xMax);
  kwsJoint->dof(1)->isBounded(true);
  kwsJoint->dof(1)->bounds(robotBounds->yMin, robotBounds->yMax);

  boundingBox->extraDof(0)->CkwsDof::bounds(-2, 2);

  SDI_F->initialized = 1;
  return ETHER;
}

/* --- GoToPosition -------------------------------------------------- */

/** walkLoadObstacleStart  -  codel START of LoadObstacle
 *
 * \return ETHER
 */

ACTIVITY_EVENT
walkLoadObstacleStart(WalkName* obstacleName, int* report)
{
  CkppKCDPolyhedronShPtr polyhedron;
  ChppciOpenHrpClient openHrpCorbaClient(hppPlanner);

  if (openHrpCorbaClient.loadObstacleModel(std::string(obstacleName->name),
					   "obstacle",
					   polyhedron) != KD_OK) {
    *report = S_walk_FAILED_TO_LOAD_OBSTACLE;
    return ETHER;
  }
  hppPlanner->addObstacle(polyhedron);

  return ETHER;
}

/* --- GoToPosition -------------------------------------------------- */

/** walkGoToPositionMain  -  codel EXEC of GoToPosition
 *
 * \return OK
 */

ACTIVITY_EVENT
walkGoToPositionMain(WalkFromToPosition *fromTo, int *report)
{
  if (!SDI_F->initialized) { *report = S_walk_NOT_INITIALIZED; return ETHER; }

  double fromRzRad = fromTo->fromPosition.RzPos * M_PI/180.;
  double toRzRad = fromTo->toPosition.RzPos * M_PI/180.;

  hppPlanner->initConfig(
    fromTo->fromPosition.xPos, fromTo->fromPosition.yPos,fromRzRad);
  hppPlanner->goalConfig(
    fromTo->toPosition.xPos, fromTo->toPosition.yPos, toRzRad);

  if (hppPlanner->solve() != KD_OK) {
    *report = S_walk_NO_SOLUTION;
    return ETHER;
  }

  return ETHER;
}


/*------------------------------------------------------------------------
 * SetUnsetHalfTurn
 *
 * Description:
 *
 * Reports:      OK
 */

/* walkSetUnsetHalfTurnMain  -  codel EXEC of SetUnsetHalfTurn
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
walkSetUnsetHalfTurnMain(int *report)
{

  if (!SDI_F->initialized) {
    *report = S_walk_NOT_INITIALIZED;
    return ETHER;
  }

  bool val =  hppPlanner->doHalfTurn() ;

  if (val == true){
    hppPlanner->doHalfTurn(false) ;
    cout << "attDoHalfTurn is now FALSE" << endl ;
  }
  else{
    hppPlanner->doHalfTurn(true) ;
    cout << "attDoHalfTurn is now TRUE" << endl ;
  }

  return ETHER;
}


/* --- GetConfig ----------------------------------------------------- */

/** walkGetConfigMain  -  codel EXEC of GetConfig
 *
 * Reports:      OK
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkGetConfigMain(int *report)
{
  double robotConfig[40];
  ghrp2_config_t cfg;
  POSTER_ID p;
  STATUS s;
  ktStatus r;
  int i;

  s = posterFind("hrp2State", &p);
  if (s != OK) { *report = S_walk_NO_MODEL; return ETHER; }

  if (posterRead(p, 0, &cfg, sizeof(cfg)) != sizeof(cfg)) {
    *report = S_walk_NO_MODEL; return ETHER;
  }

  for(i=0; i<40; i++) {
    robotConfig[i] = cfg.angles[i];
  }

  r = hppPlanner->initOpenHRPPosturePatGenRad(robotConfig);
  if (r != KD_OK) {
    *report = S_walk_NO_SOLUTION;
    return ETHER;
  }

  return ETHER;
}


/* --- SavePath ------------------------------------------------------ */

/** walkSavePathMain  -  codel EXEC of SavePath
 *
 * Reports:      OK
 *              S_walk_NO_SUCH_FILE
 *
 * \returns EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
walkSavePathMain(WalkName *name, int *report)
{
  FILE *fpos, *frpy, *fzmp;
  char full[1024];
  ktStatus s;

  vector<CkwsConfigShPtr> path;
  vector<double> walkTime;
  vector< vector<double> > zmp;
  vector< vector<double> > waist;
  vector<double> ohrp(40);
  unsigned int walkIndex, i;

  /* get data */
  s = hppPlanner->solution(path, walkTime, zmp, waist);
  if (path.size() < 1) return ETHER;

  /* open pos */
  sprintf(full, "%s.pos", name->name);
  fpos = fopen(full, "w");
  if (!fpos) {
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  /* open rpy */
  sprintf(full, "%s.rpy", name->name);
  frpy = fopen(full, "w");
  if (!frpy) {
    fclose(fpos);
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  /* open zmp */
  sprintf(full, "%s.zmp", name->name);
  fzmp = fopen(full, "w");
  if (!fzmp) {
    fclose(fpos); fclose(frpy);
    *report = S_walk_NO_SUCH_FILE; return ETHER;
  }

  /* write data */
  for (walkIndex = 0; walkIndex < walkTime.size(); walkIndex++) {
    fprintf(fpos, "%g ", walkTime[walkIndex]);
    fprintf(frpy, "%g ", walkTime[walkIndex]);
    fprintf(fzmp, "%g ", walkTime[walkIndex]);

    toOpenHRPConfig(path[walkIndex], ohrp);
    for(i = 0; i<ohrp.size(); i++) fprintf(fpos, "%g ", ohrp[i]);
    fprintf(fpos, "\n");

    fprintf(frpy, "%g %g %g\n",
	    waist[walkIndex][3], waist[walkIndex][4], waist[walkIndex][5]);

    fprintf(fzmp, "%g %g %g\n",
	    zmp[walkIndex][0], zmp[walkIndex][1], zmp[walkIndex][2]);
  }

  /* done */
  fclose(fpos); fclose(frpy); fclose(fzmp);
  return ETHER;
}


/* --- ExecutePath --------------------------------------------------- */

/** walkExecutePathStart  -  codel START of ExecutePath
 *
 * \return OK
 */

static vector<CkwsConfigShPtr> path;
static vector<double> walkTime;
static vector< vector<double> > zmp;
static vector< vector<double> > waist;
static unsigned int walkIndex;

ACTIVITY_EVENT
walkExecutePathStart(int *report)
{
  if (!SDI_F->initialized) { *report = S_walk_NOT_INITIALIZED; return ETHER; }

  ktStatus s;

  s = hppPlanner->solution(path, walkTime, zmp, waist);
  walkIndex = 0;
  if (path.size() < 1) return ETHER;
  return EXEC;
}

/** walkExecutePathMain  -  codel EXEC of ExecutePath
 *
 * \return OK
 */

ACTIVITY_EVENT
walkExecutePathMain(int *report)
{
  int i, n;
  vector<double> ohrp;
  ghrp2_config_t cfg;

  ohrp.resize(40);

  /* flood fill the buffer */
  for (walkIndex = 0; walkIndex < walkTime.size(); walkIndex++) {
    cfg.time = walkTime[walkIndex];

    toOpenHRPConfig(path[walkIndex], ohrp);
    for(i = 0; i<GHRP2_MAXDOF; i++) cfg.angles[i] = ohrp[i];

    cfg.zmp[0] = zmp[walkIndex][0];
    cfg.zmp[1] = zmp[walkIndex][1];
    cfg.zmp[2] = zmp[walkIndex][2];

    cfg.waistRpy[0] = waist[walkIndex][3];
    cfg.waistRpy[1] = waist[walkIndex][4];
    cfg.waistRpy[2] = waist[walkIndex][5];

    n = netshring_push(shtarget, &cfg, sizeof(cfg));
    if (n < 0 || n != sizeof(cfg)) {
      *report = S_walk_CANNOT_PUSH;
      return ETHER;
    }

    if (walkIndex % 500 == 0) printf("wrote %g\n", cfg.time);
  }

  printf("wrote config up to time %g\n", cfg.time);
  return ETHER;
}
